package com.fundi.framework.common.displaylist;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.Serializable;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.StringTokenizer;
import java.util.regex.Pattern;

import javax.xml.bind.annotation.XmlRootElement;

import com.fundi.framework.common.cslcomms.CSLObject;
import com.fundi.framework.common.displaylist.ResourceDefn.ResourceType;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.NumberEditor;

@XmlRootElement
public class PropDefn implements Serializable {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private static final long serialVersionUID = 1L;
	
	public static String displayFormatCharacter = "CHAR"; //$NON-NLS-1$
	public static String displayFormatInteger = "INT"; //$NON-NLS-1$
	public static String displayFormatList = "LIST"; //$NON-NLS-1$
	public static String displayFormatDate = "DATE"; //$NON-NLS-1$
	public static String displayFormatBoolean = "BOOL"; //$NON-NLS-1$
	
	public static int defaultLength = 0;
	public static boolean defaultRequired = false;
	public static boolean defaultEditable = true;
	public static boolean defaultKey = false;
	public static boolean defaultFormHide = false;
	
	public enum EditResult {
		OK(Messages.getString("PropDefn.0")), //$NON-NLS-1$
		TOO_LONG(Messages.getString("PropDefn.1")), //$NON-NLS-1$
		VALUE_REQUIRED(Messages.getString("PropDefn.2")), //$NON-NLS-1$
		INVALID_BOOLEAN(Messages.getString("PropDefn.3")), //$NON-NLS-1$
		INVALID_LIST_VALUE(Messages.getString("PropDefn.4")), //$NON-NLS-1$
		NOT_AN_INTEGER(Messages.getString("PropDefn.5")), //$NON-NLS-1$
		TOO_MANY_DIGITS(Messages.getString("PropDefn.6")), //$NON-NLS-1$
		NOT_IN_A_RANGE(Messages.getString("PropDefn.7")), //$NON-NLS-1$
		BAD_DATE_FORMAT(Messages.getString("PropDefn.8")), //$NON-NLS-1$
		INVALID_NAME(Messages.getString("PropDefn_9")), //$NON-NLS-1$
		INVALID_NAME_ALPHA_START(Messages.getString("PropDefn_10")); //$NON-NLS-1$

		private final String message;
		
		private EditResult(String message) {
			this.message = message;
		}
		
		public String toString() {
			return message;
		}
	}
	
	private String propsName = ""; //$NON-NLS-1$	//resourceDefn name
	private String keyLabel = ""; //$NON-NLS-1$
	private String displayName = ""; //$NON-NLS-1$
	private String cmdLiteral = ""; //$NON-NLS-1$
	private String cmdContext = ""; //$NON-NLS-1$
	private String displayFormat = ""; //$NON-NLS-1$
	private String tip = ""; //$NON-NLS-1$
	private int length = defaultLength;
	private String fieldValidation = ""; //$NON-NLS-1$
	private String defaultValue = ""; //$NON-NLS-1$
	private boolean required = defaultRequired;
	private boolean editable = defaultEditable;
	private boolean key = defaultKey;
	private boolean formHide = defaultFormHide;
	private String	group = "";		//not supported by the product as yet, but this mechanism already in use by the GUI //$NON-NLS-1$
	
	private String valueTrue = ""; //$NON-NLS-1$
	private String valueFalse = ""; //$NON-NLS-1$
	
	private ArrayList<String> ranges = new ArrayList<String>();
	private int[][] intRanges = new int[0][0];
	
	public final static String	ResourceGroup	= "ResourceGroup"; //$NON-NLS-1$
	public final static String	RgType			= "RgType";	 //$NON-NLS-1$
	public final static String	MbrName			= "MbrName";	 //$NON-NLS-1$
	public final static String	RepositoryName	= "RepositoryName";	 //$NON-NLS-1$
	public final static String	ModelName		= "ModelName";	 //$NON-NLS-1$
	public final static String	ModelType		= "ModelType";	 //$NON-NLS-1$
	public final static String	ResourceType	= "ResourceType";	 //$NON-NLS-1$
	public final static String	Notes			= "Notes";	 //$NON-NLS-1$
	
	public enum CommandContext {
		NAME("NAME"), //$NON-NLS-1$
		SET("SET"), //$NON-NLS-1$
		ROUTE("ROUTE"), //$NON-NLS-1$
		XSET("XSET"), //$NON-NLS-1$
		XROUTE("XROUTE"); //$NON-NLS-1$
		//
		private final String contextName;
		//
		private CommandContext(String contextName) {
			this.contextName = contextName;
		}
		public String getContextName() {
			return contextName;
		}
	}
	
	public static CommandContext getCommandContext(String contextName) {
		for (CommandContext cc: CommandContext.values()) {
			if (cc.contextName.equals(contextName)) {
				return cc;
			}
		}
		
		return null;
	}
	
	public enum ValidationStrategy {
		NAME {
			private final Pattern	namePattern = Pattern.compile(Messages.getString("PropDefn_0")); //$NON-NLS-1$

			@Override
			public EditResult validateName(String name) {
				if (namePattern.matcher(name.trim()).matches()) {
					return EditResult.OK;
				} else {
					return EditResult.INVALID_NAME;
				}
			}
		},
		NAME_ALPHA_START {
			private final Pattern	namePattern = Pattern.compile(Messages.getString("PropDefn_1")); //$NON-NLS-1$
			
			@Override
			public EditResult validateName(String name) {
				if (namePattern.matcher(name.trim()).matches()) {
					return EditResult.OK;
				} else {
					return EditResult.INVALID_NAME_ALPHA_START;
				}
			}
		};
		//
		public abstract PropDefn.EditResult validateName(String name);
	}
	
	@Override
	public String toString() {
		StringBuilder	sb = new StringBuilder();
		
		sb.append(keyLabel);
		sb.append(" ("); //$NON-NLS-1$
		sb.append(displayName);
		sb.append(")"); //$NON-NLS-1$
		if (isKey()) {
			sb.append(", key"); //$NON-NLS-1$
		}
		if (isEditable()) {
			sb.append(", editable"); //$NON-NLS-1$
		}
		if (isRequired()) {
			sb.append(", required"); //$NON-NLS-1$
		}
		
		return sb.toString();
	}

	public PropDefn() {
	
		super();
	}
	
	public PropDefn(CSLObject cslObj) {
		
		initialise(cslObj);
	}

	private void initialise(CSLObject cslObj) {
		
		NumberEditor edt;
		
		propsName = cslObj.getFieldValueByLlbl("objectName"); //$NON-NLS-1$
		keyLabel = cslObj.getFieldValueByLlbl("keyLabel"); //$NON-NLS-1$
		displayName = cslObj.getFieldValueByLlbl("displayName"); //$NON-NLS-1$
		cmdLiteral = cslObj.getFieldValueByLlbl("cmdLiteral"); //$NON-NLS-1$
		cmdContext = cslObj.getFieldValueByLlbl("cmdContext"); //$NON-NLS-1$
		displayFormat = cslObj.getFieldValueByLlbl("displayFormat"); //$NON-NLS-1$
		tip = cslObj.getFieldValueByLlbl("tip"); //$NON-NLS-1$
		edt = new NumberEditor(cslObj.getFieldValueByLlbl("length")); //$NON-NLS-1$
		if (edt.isValidInteger(5)) {
			length = edt.intValue();
		}
		else {
			if (cslObj.getFieldValueByLlbl("length").equals("*")) { //$NON-NLS-1$ //$NON-NLS-2$
				length = defaultLength;
			}
		}
		fieldValidation = cslObj.getFieldValueByLlbl("fieldValidation"); //$NON-NLS-1$
		setRanges();
		defaultValue = cslObj.getFieldValueByLlbl("default"); //$NON-NLS-1$
		required = cslObj.getFieldValueByLlbl("required").equalsIgnoreCase("y"); //$NON-NLS-1$ //$NON-NLS-2$
		editable = cslObj.getFieldValueByLlbl("editable").equalsIgnoreCase("y"); //$NON-NLS-1$ //$NON-NLS-2$
		key = cslObj.getFieldValueByLlbl("editable").equalsIgnoreCase("k"); //$NON-NLS-1$ //$NON-NLS-2$
		formHide = cslObj.getFieldValueByLlbl("formHide").equalsIgnoreCase("y"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	private void setRanges() {
		
		StringTokenizer parser;
		String rangeSeparator = ","; //$NON-NLS-1$
		String minMaxSeparator = "-";  //$NON-NLS-1$
		int idx;
		String range;
		int sepPos;
		
		if (fieldValidation.length() == 0 || displayFormat.equals(displayFormatCharacter) || displayFormat.equals(displayFormatDate)) {
			return;
		}
		
		parser = new StringTokenizer(fieldValidation, rangeSeparator);
		while (parser.hasMoreTokens()) {
			ranges.add(parser.nextToken().trim());
		}
		
		//If a set of integer ranges, build list of min./max. values;
		if (displayFormat.equals(displayFormatInteger)) {
			intRanges = new int[ranges.size()][2];
			for (idx = 0; idx < ranges.size(); idx++) {
				range = ranges.get(idx);
				sepPos = range.indexOf(minMaxSeparator);
				if (sepPos < 0) {
					intRanges[idx][0] = Integer.parseInt(range);
					intRanges[idx][1] = intRanges[idx][0];
				}
				else {
					intRanges[idx][0] = Integer.parseInt(range.substring(0, sepPos));
					intRanges[idx][1] = Integer.parseInt(range.substring(sepPos + 1));
				}
			}
		}
		
		//If a boolean property, populate true/false fields
		if (displayFormat.equals(displayFormatBoolean) && ranges.size() == 2) {
			valueTrue = ranges.get(0);
			valueFalse = ranges.get(1);
		}
	}

	public String getPropsName() {
		return propsName;
	}

	public String getKeyLabel() {
		return keyLabel;
	}

	public String getDisplayName() {
		return displayName;
	}

	public String getCmdLiteral() {
		return cmdLiteral;
	}

	public String getCmdContext() {
		return cmdContext;
	}

	public String getDisplayFormat() {
		return displayFormat;
	}

	public String getTip() {
		return tip;
	}

	public int getLength() {
		return length;
	}

	public String getFieldValidation() {
		return fieldValidation;
	}

	public String getDefaultValue() {
		return defaultValue;
	}

	public boolean isRequired() {
		return required;
	}

	public boolean isEditable() {
		return editable;
	}

	public boolean isKey() {
		return key;
	}

	public boolean isFormHide() {
		return formHide;
	}

	public String getValueTrue() {
		return valueTrue;
	}

	public String getValueFalse() {
		return valueFalse;
	}
	
	public EditResult edit(String value) {
		
		NumberEditor numEdt;
		int idx;
		SimpleDateFormat dateFmt;
		ParsePosition pos;
		Date parseDate;
		ResourceType rt;
		
		value = value.trim();
		
		if (value.length() == 0 && defaultValue.length() > 0) {
			value = defaultValue;
		}
		
		if (value.trim().length() == 0) {
			if (isRequired()) { 
				return EditResult.VALUE_REQUIRED;
			} else {
				return EditResult.OK;
			}
		}
		
		rt = ResourceDefn.getResourceType(propsName); 
		if ("NAME".equals(cmdLiteral) && rt != null) { //$NON-NLS-1$
			//override and use specific naming rules coded into ResourceDefn
			return rt.validateName(value);
		} else if (displayFormat.equals(displayFormatBoolean)) {
			if (!ranges.contains(value)) {
				return EditResult.INVALID_BOOLEAN;
			}
		} else if (displayFormat.equals(displayFormatList)) {
			if (!value.trim().isEmpty() && !ranges.contains(value)) {	//if we have made it to here then the value is not required
				return EditResult.INVALID_LIST_VALUE;
			}
		} else if (displayFormat.equals(displayFormatInteger)) {
			numEdt = new NumberEditor(value);
			if (!numEdt.isValidInteger(20)) {  //Arbitrarily set very high number of max. digits.
				return EditResult.NOT_AN_INTEGER;
			}
			else {
				if (!numEdt.isValidInteger(length)) {  //Now test for required length.
					return EditResult.TOO_MANY_DIGITS;
				}
			}
			if (ranges.size() > 0) {
				for (idx = 0; idx < ranges.size(); idx++) {
					if (numEdt.isInRange(intRanges[idx][0], intRanges[idx][1])) {
						return EditResult.OK;
					}
				}
				return EditResult.NOT_IN_A_RANGE;
			}
		} else if (displayFormat.equals(displayFormatDate)) {
			 dateFmt = new SimpleDateFormat (fieldValidation);
			 pos = new ParsePosition(0);
			 parseDate = dateFmt.parse(value, pos);
			 if (parseDate == null) {
				 return EditResult.BAD_DATE_FORMAT;
			 }
			 value = dateFmt.format(parseDate);  //To ensure it is fully formatted into the required format.	//using if-else now because the value is actually getting changed here.
		} else if (length > 0 && value.length() > length) {
			return EditResult.TOO_LONG;
		}
		
		return EditResult.OK;
	}
	
	public String getBooleanRepresentation(boolean val) {
		
		if (val) {
			return valueTrue;
		}
		return valueFalse;
	}
	
	public List<String> getValidationListItems() {
		List<String> items;
		
		items = new ArrayList<String>();
		if ((displayFormatList.equals(displayFormat) || displayFormatBoolean.equals(displayFormat)) && fieldValidation != null) {
			items.addAll(Arrays.asList(fieldValidation.split("\\s*,\\s*"))); //$NON-NLS-1$
		}
		
		return items;
	}

	public String getGroup() {
		return group;
	}

	public void setGroup(String group) {
		this.group = group;
	}

	public void setPropsName(String propsName) {
		this.propsName = propsName;
	}

	public void setKeyLabel(String keyLabel) {
		this.keyLabel = keyLabel;
	}

	public void setDisplayName(String displayName) {
		this.displayName = displayName;
	}

	public void setCmdContext(String cmdContext) {
		this.cmdContext = cmdContext;
	}

	public void setDisplayFormat(String displayFormat) {
		this.displayFormat = displayFormat;
	}

	public void setTip(String tip) {
		this.tip = tip;
	}

	public void setLength(int length) {
		this.length = length;
	}

	public void setFieldValidation(String fieldValidation) {
		this.fieldValidation = fieldValidation;
	}

	public void setRequired(boolean required) {
		this.required = required;
	}

	public void setEditable(boolean editable) {
		this.editable = editable;
	}

	public void setKey(boolean key) {
		this.key = key;
	}

	public void setFormHide(boolean formHide) {
		this.formHide = formHide;
	}
	
}
