package com.fundi.framework.common.helpers;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.fundi.framework.common.displaylist.DisplayList;
import com.fundi.framework.common.displaylist.ListField;
import com.fundi.framework.common.displaylist.ListFieldDefn;
import com.fundi.framework.common.displaylist.ListRow;
import com.fundi.framework.common.instance.AppInstance;

public class ExportHelper {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private AppInstance instance;
	private PrintWriter writer;
	
	
	public ExportHelper(AppInstance instance) {
		
		super();
		
		initialise(instance, null);
	}
	
	public ExportHelper(AppInstance instance, PrintWriter writer) {

		super();
		
		initialise(instance, writer);
	}
	
	private void initialise(AppInstance instance, PrintWriter writer) {
		
		this.instance = instance;
		this.writer = writer;
	}
	
	public boolean export(Object source, String destPath) {
		
		return export(source, destPath, null);
	}
	
	public boolean export(Object source, String destPath, ListRow[] selectedRows) {
		
		boolean ok = false;

		try {
			writer = new PrintWriter(new File(destPath));
			ok = export(source, selectedRows);
			writer.flush();
			writer.close();
		} 
		catch (FileNotFoundException e) {
			ok = false;
			instance.addConsoleMessage(e);
		}
		
		return ok;
	}
	
	public boolean export(Object source, PrintWriter writer) {
		
		return export(source, writer, null);
	}
	
	public boolean export(Object source, PrintWriter writer, ListRow[] selectedRows) {
		
		this.writer = writer;
		
		return export(source, selectedRows);
	}
	
	public boolean export(Object source) {

		ListRow[] selectedRows = new ListRow[]{};
		
		return export(source, selectedRows);
	}
	
	public boolean export(Object source, ListRow[] selectedRows) {
		
		boolean ok = true;
		
		if (DisplayList.class.isInstance(source)) {
			ok = writeDisplayList(source, selectedRows, new FormatCSV());
		}
		
		return ok;
	}
	
	private boolean writeDisplayList(Object source, ListRow[] selectedRows, Format format) {
		
		ListRow[] rows = new ListRow[]{};
		
		if (selectedRows == null) {
			rows = ((DisplayList) source).retrieveVisibleRows().toArray(new ListRow[((DisplayList) source).getVisibleRowCount()]);
		} else {
			if (selectedRows.length == 0) {
				rows = ((DisplayList) source).retrieveVisibleRows().toArray(new ListRow[((DisplayList) source).getVisibleRowCount()]);
			} else {
				rows = selectedRows;
			}
		}
		format.preamble((DisplayList) source);
		for (ListRow row: rows) {
			format.listrow(row);
		}
		format.finish();
		
		return !(writer.checkError());
	}
	
	
	
	//CLASS////////////////////////////////////////////////////////////////////
	public abstract class Format {
		public abstract void preamble(DisplayList list);
		public abstract void listrow(ListRow row);
		public abstract void finish();
	}
	
	/**
	 * Implementation of CSV format to https://tools.ietf.org/html/rfc4180
	 * This implementation goes slightly further than the RFC minimum requirement for enclosing fields in double quotes ("):
	 * 
	 * | 6.  Fields containing line breaks (CRLF), double quotes, and commas
	 * |     should be enclosed in double-quotes.       
	 * 
	 * This implementation encloses in double quotes if any of CRLF, CR, LF, ", comma are found as I feel this makes the
	 * format more tolerant of 'less specific' parsing of new line markers by parsing software.
	 */
	public class FormatCSV extends Format {
		private final Pattern requiresQuote = Pattern.compile("\\\"|\r\n|[\r\n,]");
		private DisplayList list;
		//
		public final String fieldDelimiter = "\""; 								//$NON-NLS-1$
		public final String fieldDelimiterEsc = "\"\""; 						//$NON-NLS-1$
		public final String fieldDelimiterPat = Pattern.quote(fieldDelimiter); 	//$NON-NLS-1$
		public final String fieldSeparator = ","; 								//$NON-NLS-1$
		public final String lineFeed = "\r\n"; 									//$NON-NLS-1$
		//
		@Override
		public void preamble(DisplayList list) {
			String separator = "";
			
			this.list = list;
			for (ListFieldDefn defn: list.getFieldDefns()) {
				writer.print(separator);
				if (defn.isVisible()) {
					format(defn.getLabel());
					separator = fieldSeparator;
				}
			}
			writer.print(lineFeed);
		}
		@Override
		public void listrow(ListRow row) {
			String separator = "";
			String value;
			ListField field;
			
			for (ListFieldDefn defn: list.getFieldDefns()) {
				writer.print(separator);
				if (defn.isVisible()) {
					value = null;
					field = row.getField(defn.getLabel());
					if (field != null) {
						value = field.getValue();
						if (value == null) {
							value = "";
						}
					}
					else {
						value = "";
					}
					format(value);
					separator = fieldSeparator;
				}
			}
			writer.print(lineFeed);
		}
		@Override
		public void finish() {
		}
		private void format(String value) {
			Matcher matcher;
			
			matcher = requiresQuote.matcher(value);
			if (matcher.find()) {
				writer.print(fieldDelimiter);
				writer.print(value.replaceAll(fieldDelimiterPat, fieldDelimiterEsc));
				writer.print(fieldDelimiter);
			} else {
				writer.print(value);
			}
		}
	}
	
	//Since I know a jira for this is already in the system, may as well prepare for it...
	public class FormatJSON extends Format {
		private DisplayList list;
		//
		@Override
		public void preamble(DisplayList list) {
			this.list = list;
		}
		@Override
		public void listrow(ListRow row) {
		}
		@Override
		public void finish() {
		}
	}
}
